
from .riff_parser import parse_chunk, ListChunkDescriptor

class WavInfoChunkReader:

    def __init__(self, f, encoding):
        self.encoding = encoding

        f.seek(0)
        parsed_chunks = parse_chunk(f)

        list_chunks = [chunk for chunk in parsed_chunks.children \
                if type(chunk) is ListChunkDescriptor]

        self.info_chunk  = next((chunk for chunk in list_chunks \
                if chunk.signature == b'INFO'), None)
        
        #: 'ICOP' Copyright
        self.copyright      = self._get_field(f,b'ICOP')
        #: 'IPRD' Product
        self.product        = self._get_field(f,b'IPRD')
        #: 'IGNR' Genre
        self.genre          = self._get_field(f,b'IGNR')
        #: 'ISBJ' Supject
        self.subject        = self._get_field(f,b'ISBJ')
        #: 'IART' Artist, composer, author
        self.artist         = self._get_field(f,b'IART')
        #: 'ICMT' Comment
        self.comment        = self._get_field(f,b'ICMT')
        #: 'ISFT' Software, encoding application
        self.software       = self._get_field(f,b'ISFT')
        #: 'ICRD' Created date
        self.created_date   = self._get_field(f,b'ICRD')
        #: 'IENG' Engineer
        self.engineer       = self._get_field(f,b'IENG')
        #: 'ITCH' Technician
        self.technician     = self._get_field(f,b'ITCH')
        #: 'IKEY' Keywords, keyword list
        self.keywords       = self._get_field(f,b'IKEY')
        #: 'INAM' Name, title
        self.title          = self._get_field(f,b'INAM')
        #: 'ISRC' Source
        self.source         = self._get_field(f,b'ISRC')
        #: 'TAPE' Tape
        self.tape           = self._get_field(f,b'TAPE')
        #: 'IARL' Archival Location
        self.archival_location = self._get_field(f,b'IARL')
        #: 'ISFT' Software
        self.software       = self._get_field(f,b'ISFT')
        #: 'ICSM' Commissioned
        self.commissioned   = self._get_field(f,b'ICMS')
        


    def _get_field(self, f, field_ident):

        search = next( ( (chunk.start, chunk.length) for chunk in self.info_chunk.children \
                if chunk.ident == field_ident ), None)

        if search is not None:
            f.seek(search[0])
            data = f.read(search[1])
            return data.decode(self.encoding).rstrip('\0')
        else:
            return None


    def to_dict(self):
        """
        A dictionary with all of the key/values read from the INFO scope.
        """
        return {'copyright':    self.copyright,
                'product':  self.product,
                'genre':    self.genre,
                'artist':   self.artist,
                'comment':  self.comment,
                'software': self.software,
                'created_date': self.created_date,
                'engineer': self.engineer,
                'keywords': self.keywords,
                'title':    self.title,
                'source':   self.source,
                'tape':     self.tape,
                'commissioned': self.commissioned,
                'software': self.software,
                'archival_location':self.archival_location,
                'subject':  self.subject,
                'technician':self.technician
                }









from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='wavinfo',
      version='1.4',
      author='Jamie Hardt',
      author_email='jamiehardt@me.com',
      description='Probe WAVE Files for iXML, Broadcast-WAVE and other metadata.',
      long_description_content_type="text/markdown",
      long_description=long_description,
      url='https://github.com/iluvcapra/wavinfo',
      project_urls={
          'Source':
              'https://github.com/iluvcapra/wavinfo',
          'Documentation':
              'https://wavinfo.readthedocs.io/',
          'Issues':
              'https://github.com/iluvcapra/wavinfo/issues',
      },
      packages=['wavinfo'],
      classifiers=['Development Status :: 5 - Production/Stable',
          'License :: OSI Approved :: MIT License',
          'Topic :: Multimedia',
          'Topic :: Multimedia :: Sound/Audio',
          "Programming Language :: Python :: 3.5",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "Programming Language :: Python :: 3.8"],
      keywords='waveform metadata audio ebu smpte avi library film tv editing editorial',
      install_requires=['lxml']
      )



#-*- coding: utf-8 -*-
import struct
import os
import sys
from collections import namedtuple

if sys.version[0] == '3':
    import pathlib
else:
    import urlparse, urllib

from .riff_parser import parse_chunk, ChunkDescriptor, ListChunkDescriptor
from .wave_ixml_reader import WavIXMLFormat
from .wave_bext_reader import WavBextReader
from .wave_info_reader import WavInfoChunkReader

#: Calculated statistics about the audio data.
WavDataDescriptor = namedtuple('WavDataDescriptor','byte_count frame_count')

#: The format of the audio samples.
WavAudioFormat = namedtuple('WavAudioFormat','audio_format channel_count sample_rate byte_rate block_align bits_per_sample')

class WavInfoReader():
    """
    Parse a WAV audio file for metadata.
    """

    def __init__(self, path, info_encoding='latin_1', bext_encoding='ascii'):
        """
        Create a new reader object.

        :param path: A filesystem path to the wav file you wish to probe.

        :param info_encoding: The text encoding of the INFO metadata fields.
          latin_1/Win CP1252 has always been a pretty good guess for this.

        :param bext_encoding: The text encoding to use when decoding the string
          fields of the Broadcast-WAV extension. Per EBU 3285 this is ASCII
          but this parameter is available to you if you encounter a werido.
        """
        absolute_path = os.path.abspath(path)

        if sys.version[0] == '3':
            #: `file://` url for the file.
            self.url = pathlib.Path(absolute_path).as_uri()
        else:
            self.url = urlparse.urljoin('file:', urllib.pathname2url(absolute_path))

        with open(path, 'rb') as f:
            chunks = parse_chunk(f)

            self.main_list = chunks.children
            f.seek(0)

            #: :class:`wavinfo.wave_reader.WavAudioFormat`
            self.fmt    = self._get_format(f)

            #: :class:`wavinfo.wave_bext_reader.WavBextReader` with Broadcast-WAV metadata
            self.bext   = self._get_bext(f, encoding=bext_encoding)

            #: :class:`wavinfo.wave_ixml_reader.WavIXMLFormat` with iXML metadata
            self.ixml   = self._get_ixml(f)

            #: :class:`wavinfo.wave_info_reader.WavInfoChunkReader` with RIFF INFO metadata
            self.info   = self._get_info(f, encoding=info_encoding)
            self.data   = self._describe_data(f)

    def _find_chunk_data(self, ident, from_stream, default_none=False):
        chunk_descriptor = None
        top_chunks = (chunk for chunk in self.main_list if type(chunk) is ChunkDescriptor)

        if default_none:
            chunk_descriptor = next((chunk for chunk in top_chunks if chunk.ident == ident),None)
        else:
            chunk_descriptor = next((chunk for chunk in top_chunks if chunk.ident == ident))

        if chunk_descriptor:
            return chunk_descriptor.read_data(from_stream)
        else:
            return None


    def _describe_data(self,f):
        data_chunk = next(c for c in self.main_list if c.ident == b'data')

        return WavDataDescriptor(byte_count= data_chunk.length,
                frame_count= int(data_chunk.length / self.fmt.block_align))


    def _get_format(self,f):
        fmt_data = self._find_chunk_data(b'fmt ',f)

        # The format chunk is
        # audio_format    U16
        # channel_count   U16
        # sample_rate     U32   Note an integer
        # byte_rate       U32   == SampleRate * NumChannels * BitsPerSample/8
        # block_align     U16   == NumChannels * BitsPerSample/8
        # bits_per_sampl  U16
        packstring = "<HHIIHH"
        rest_starts = struct.calcsize(packstring)

        unpacked = struct.unpack(packstring, fmt_data[:rest_starts])

        #0x0001	WAVE_FORMAT_PCM	PCM
        #0x0003	WAVE_FORMAT_IEEE_FLOAT	IEEE float
        #0x0006	WAVE_FORMAT_ALAW	8-bit ITU-T G.711 A-law
        #0x0007	WAVE_FORMAT_MULAW	8-bit ITU-T G.711 µ-law
        #0xFFFE	WAVE_FORMAT_EXTENSIBLE	Determined by SubFormat

        #https://sno.phy.queensu.ca/~phil/exiftool/TagNames/RIFF.html
        return WavAudioFormat(audio_format = unpacked[0],
                    channel_count   = unpacked[1],
                    sample_rate     = unpacked[2],
                    byte_rate       = unpacked[3],
                    block_align     = unpacked[4],
                    bits_per_sample = unpacked[5]
                    )

    def _get_info(self, f, encoding):
        finder = (chunk.signature for chunk in self.main_list \
                if type(chunk) is ListChunkDescriptor)

        if b'INFO' in finder:
            return WavInfoChunkReader(f, encoding)

    def _get_bext(self, f, encoding):
        bext_data = self._find_chunk_data(b'bext',f,default_none=True)
        if bext_data:
            return WavBextReader(bext_data, encoding)
        else:
            return None

    def _get_ixml(self,f):
        ixml_data = self._find_chunk_data(b'iXML',f,default_none=True)
        if ixml_data is None:
            return None

        ixml_string = ixml_data.rstrip(b'\0')
        return WavIXMLFormat(ixml_string)

    def walk(self):
        """
        Walk all of the available metadata fields.

        :yields: a string, the :scope: of the metadatum, the string :name: of the
        metadata field, and the value.
        """

        scopes = ('fmt','data')#,'bext','ixml','info')

        for scope in scopes:
            attr = self.__getattribute__(scope)
            for field in attr._fields:
                yield scope, field, attr.__getattribute__(field)



import struct

class WavBextReader:
    def __init__(self,bext_data,encoding):
        """
        Read Broadcast-WAV extended metadata.
        :param best_data: The bytes-like data.
        "param encoding: The encoding to use when decoding the text fields of the
                 BEXT metadata scope. According to EBU Rec 3285 this shall be ASCII.
        """
        packstring = "<256s"+ "32s" + "32s" + "10s" + "8s" + "QH" + "64s" + "hhhhh" + "180s"

        rest_starts = struct.calcsize(packstring)
        unpacked = struct.unpack(packstring, bext_data[:rest_starts])

        def sanatize_bytes(bytes):
            first_null = next( (index for index, byte in enumerate(bytes) if byte == 0 ), None )
            if first_null is not None:
                trimmed = bytes[:first_null]
            else:
                trimmed = bytes

            decoded = trimmed.decode(encoding)
            return decoded

        #: Description. A free-text field up to 256 characters long.
        self.description     = sanatize_bytes(unpacked[0])
        #: Originator. Usually the name of the encoding application, sometimes
        #: a artist name.
        self.originator      = sanatize_bytes(unpacked[1])
        #: A unique identifer for the file, a serial number.
        self.originator_ref  = sanatize_bytes(unpacked[2])
        #: Date of the recording, in the format YYY-MM-DD
        self.originator_date = sanatize_bytes(unpacked[3])
        #: Time of the recording, in the format HH:MM:SS.
        self.originator_time = sanatize_bytes(unpacked[4])
        #: The sample offset of the start of the file relative to an
        #: epoch, usually midnight the day of the recording. 
        self.time_reference  = unpacked[5]
        #: A variable-length text field containing a list of processes and
        #: and conversions performed on the file.
        self.coding_history  = sanatize_bytes(bext_data[rest_starts:])
        #: BEXT version. 
        self.version         = unpacked[6]
        #: SMPTE 330M UMID of this audio file, 64 bytes are allocated though the UMID
        #: may only be 32 bytes long.
        self.umid            = None
        #: EBU R128 Integrated loudness, in LUFS.
        self.loudness_value          = None
        #: EBU R128 Loudness rante, in LUFS.
        self.loudness_range          = None
        #: True peak level, in dBFS TP
        self.max_true_peak           = None
        #: EBU R128 Maximum momentary loudness, in LUFS
        self.max_momentary_loudness  = None
        #: EBU R128 Maximum short-term loudness, in LUFS.
        self.max_shortterm_loudness  = None

        if self.version > 0:
            self.umid = unpacked[7]

        if self.version > 1:
            self.loudness_value          = unpacked[8] / 100.0
            self.loudness_range          = unpacked[9] / 100.0
            self.max_true_peak           = unpacked[10] / 100.0
            self.max_momentary_loudness  = unpacked[11] / 100.0
            self.max_shortterm_loudness  = unpacked[12] / 100.0


    def to_dict(self):
        return {'description':      self.description,
                'originator':       self.originator,
                'originator_ref':   self.originator_ref,
                'originator_date':  self.originator_date,
                'originator_time':  self.originator_time,
                'time_reference':   self.time_reference,
                'version':          self.version,
                'coding_history':   self.coding_history,
                'loudness_value':   self.loudness_value,
                'loudness_range':   self.loudness_range,
                'max_true_peak':    self.max_true_peak,
                'max_momentary_loudness':   self.max_momentary_loudness,
                'max_shortterm_loudness':   self.max_shortterm_loudness
                }

